import json

import numpy as np
from dataclasses import dataclass, field

@dataclass
class Invocation:
   index : int

@dataclass
class Source(Invocation):
   value : 'typing.Any'

@dataclass
class StartFlow(Invocation):
   pass

@dataclass
class Sink(Invocation):
   merge: bool = False

@dataclass
class InvokeTask(Invocation):
   name: str
   parameters : 'typing.Any' = None
   base: str = None
   doc: str = None
   merge: bool = False

@dataclass
class InvokeFlow(Invocation):
   merge: bool = False


class Flow:

   __classes = {
      'Source' : Source,
      'Sink' : Sink,
      'InvokeTask' : InvokeTask,
      'InvokeFlow' : InvokeFlow,
      'StartFlow' : StartFlow
   }

   def __init__(self,size=0,serialized=None,name=None):
      self._name = name
      if serialized is None:
         assert size>0
         self._F = np.zeros((size,size),dtype=int)
         self._tasks = [None]*size
      else:
         self._F = np.array(serialized['F'])
         self._tasks = [Flow.__classes[T[0]](**T[1]) for T in serialized['T']]

   @property
   def name(self):
      return self._name

   @name.setter
   def name(self,value):
      self._name = value

   @property
   def F(self):
      return self._F

   def __getitem__(self,index):
      return self._tasks[index]

   def __setitem__(self,index,value):
      self._tasks[index] = value

   def __len__(self):
      return len(self._tasks)

   def __str__(self):
      return json.dumps(self.save())

   def save(self):
      R = {
         'F' : self._F.tolist(),
         'T' : [[T.__class__.__name__,T.__dict__] for T in self._tasks]
      }
      return R



from littleflow import Parser, Compiler, Runner, Context, FunctionTaskContext

class WorkflowFailure(Exception):
   pass

class TaskFailure(WorkflowFailure):
   def __init__(self, message, *args, task=None, **kwargs):
      if task is None:
         super().__init__(message,*args,**kwargs)
      else:
         super().__init__(f'Task {task} failed: {message}',*args,**kwargs)


def run_workflow(workflow,lookup=None,context=None,input=None):
   p = Parser()
   c = Compiler()
   model = p.parse(workflow)
   flow = c.compile(model)

   runner = Runner()
   if context is None:
      context = Context(flow,task_context=FunctionTaskContext(lookup))
   elif lookup is not None:
      context.task_context = FunctionTaskContext(lookup)

   try:
      runner.start(context,input=input)

      while not context.ending.empty():
         runner.next(context,context.ending.get())
   except TaskFailure as ex:
      raise WorkflowFailure('The workflow failed due to a task failure.') from ex
   except Exception as ex:
      raise WorkflowFailure('The workflow failed due to an exception.') from ex


   return context.input_for(len(flow)-1), context



import logging
from lark import Lark, Token, Tree

from .model import Workflow, Declaration, SubFlow, Statement, Start, End, Iterate, Task, LiteralSource, ResourceSource, ResourceSink, ParameterLiteral, LiteralType

grammar = r"""
flow: (declaration | flow_statement)+
declaration: DECLARE NAME (EQUAL NAME)? parameter_literal? doc_comment? ";"?
flow_statement: (source ARROW)? step (ARROW step)* (ARROW destination)? ";"?
step: LABEL? (STAR | MERGE)? (task_list | subflow | conditional) LABEL?
task_list : task (OR task)*
task: NAME parameter_literal?
subflow: "{" flow_statement+ "}"
conditional: "if" EXPR "then" step ("elif" EXPR "then" step)* ("else" step)?
source: LABEL | resource | resource_literal
destination: MERGE? (LABEL | resource)
resource: URI parameter_literal?
parameter_literal: empty_parameter | json_object_parameter | json_array_parameter | yaml_parameter
empty_parameter: "(" ")"
json_object_parameter: "({" JSON_OBJECT_PARAMETER_BODY? "})"
json_array_parameter: "([" JSON_ARRAY_PARAMETER_BODY? "])"
yaml_parameter: "(-" YAML_PARAMETER_BODY? "-)"
resource_literal : empty_resource | json_object_resource | json_array_resource | yaml_resource
empty_resource: "<" ">"
json_object_resource: "<{" JSON_OBJECT_RESOURCE_BODY? "}>"
json_array_resource: "<[" JSON_ARRAY_RESOURCE_BODY? "]>"
yaml_resource: "<-" YAML_RESOURCE_BODY? "->"
doc_comment: "'''" DOC_COMMENT_BODY_SINGLE? "'''" | "\"\"\"" DOC_COMMENT_BODY_DOUBLE? "\"\"\""
ARROW: "→" | "->"
STAR: "*"
OR: "|"
MERGE: ">"
EQUAL: "="
DECLARE: "@" NAME
LABEL: ":" NAME
NAME: /[a-zA-Z_][\w\-_:]*/
DEC_NUMBER: /0|[1-9][\d_]*/i
EXPR: /`[^`]*`/
STRING: /("(?!"").*?(?<!\\)(\\\\)*?"|'(?!'').*?(?<!\\)(\\\\)*?')/i
URI: /<[^>\-{[][^>]*>/
JSON_OBJECT_PARAMETER_BODY: /([^}]+|(}[^)]))+/
JSON_ARRAY_PARAMETER_BODY: /([^]]+|(][^)]))+/
YAML_PARAMETER_BODY: /([^-]+|(-[^)]))+/
JSON_OBJECT_RESOURCE_BODY: /([^}]+|(}[^>]))+/
JSON_ARRAY_RESOURCE_BODY: /([^]]+|(][^>]))+/
YAML_RESOURCE_BODY: /([^-]+|(-[^>]))+/
DOC_COMMENT_BODY_SINGLE: /([^']+|'[^']|''[^'])+/
DOC_COMMENT_BODY_DOUBLE: /([^"]+|"[^"]|""[^"])+/
COMMENT: /#[^\n]*/
%import common.WS
%ignore WS
%ignore COMMENT
"""

def iter_tree(top):
   context = [(True,top,[child for child in top.children if isinstance(child,Tree)])]
   while len(context)>0:
      first, current, subtrees = context.pop()

      if not first:
         yield first, current
         continue

      yield first, current

      context.append((False,current,[]))
      subtrees.reverse()
      for subtree in subtrees:
         context.append((True,subtree,[child for child in subtree.children if isinstance(child,Tree)]))

def find_position(tree,target):
   for item in tree.iter_subtrees():
      for child in item.children:
         if isinstance(child,Token):
            target.line = child.line
            target.column = child.column
            return;
class Parser:

   def __init__(self):
      self._parser = Lark(grammar,parser='lalr',start='flow',propagate_positions=True)

   def parse(self,source):
      if type(source)!=str:
         source = source.read()
      ast = self._parser.parse(source)

      from_start = []

      workflow = Workflow()

      flow = None
      statement = None
      iterate = False
      subject = workflow
      source = None
      media_tyope = None
      input_label = None
      output_label = None
      source_labeled = False
      last_step = None
      ancestors = []

      def realize_step(step):
         nonlocal source_labeled, last_step
         last_step = step
         if statement is not None:
            statement.steps.append(step)
         workflow.indexed.append(step)
         if not source_labeled and statement.source is not None:
            flow.named_inputs[statement.source] = step
            source_labeled = True
         if input_label is not None:
            label = input_label.value[1:]
            if label in ['start','end']:
               raise ValueError(f'{input_label.line}:{input_label.column} {label} is a reserved label')
            if label in flow.named_inputs:
               raise ValueError(f'{input_label.line}:{input_label.column} input label {label} already exists')
            flow.named_inputs[label] = step
         if output_label is not None:
            label = output_label.value[1:]
            if label in ['start','end']:
               raise ValueError(f'{output_label.line}:{output_label.column} {label} is a reserved label')
            if label in flow.named_outputs:
               raise ValueError(f'{output_label.line}:{output_label.column} output label {label} already exists')
            flow.named_outputs[label] = step

      for at_start, item in iter_tree(ast):
         if not at_start:
            if item.data=='subflow':
               index += 1
               end.index = index
               workflow.indexed.append(end)
               subject = flow

               flow, statement = ancestors.pop()
               end = flow.named_inputs['end']
               start = flow.named_outputs['start']
            elif item.data=='parameter_literal':
               if subject is not None:
                  subject.parameters = literal
               else:
                  print(item)
                  assert False, f'{line}:{column} Unknown context for parameter literal'
            elif item.data=='flow_statement':
               if statement.destination is not None:
                  flow.named_outputs[statement.destination] = last_step

            continue

         if item.data=='flow':
            index = 0
            flow = SubFlow(index)
            workflow.flows.append(flow)
            start = Start(index)
            workflow.indexed.append(start)

            flow.named_outputs['start'] = start

            end = End(-1)
            flow.named_inputs['end'] = end
            flow.named_inputs['start'] = start
         elif item.data=='subflow':
            ancestors.append((flow,statement))
            index += 1
            flow = SubFlow(index,merge=merge)
            workflow.flows.append(flow)
            start = Start(index)
            if iterate:
               step = Iterate(flow)
               step.index = step.step.index
               realize_step(step)
            else:
               realize_step(flow)
            flow.named_outputs['start'] = start
            end = End(-1)
            flow.named_inputs['end'] = end
            flow.named_inputs['start'] = start
         elif item.data=='flow_statement':
            iterate = False
            step = None
            source = None
            source_labeled = False
            statement = Statement()
            find_position(item,statement)
            flow.statements.append(statement)
         elif item.data=='step':
            subject = None

            input_label = None
            output_label = None
            before = True
            iterate = False
            merge = False
            for child in item.children:
               if isinstance(child,Token) and child.type=='STAR':
                  iterate = True
               elif isinstance(child,Token) and child.type=='MERGE':
                  merge = True
               elif isinstance(child,Token) and child.type=='LABEL':
                  if before:
                     input_label = child
                  else:
                     output_label = child
               elif not isinstance(child,Token):
                  before = False
         elif item.data=='source':
            if isinstance(item.children[0],Token) and item.children[0].type=='LABEL':
               statement.source = item.children[0].value[1:]
         elif item.data=='resource':
            index += 1
            subject = ResourceSource(index,item.children[0].value[1:-1])
            realize_step(subject)
         elif item.data=='resource_literal':
            index += 1
            literal = LiteralSource(index,None)
            realize_step(literal)
         elif item.data=='empty_resource':
            literal.line = item.meta.line
            literal.column = item.meta.column
         elif item.data=='yaml_resource':
            literal.value = item.children[0].value if len(item.children)>0 else ''
            literal.type = LiteralType.YAML
            literal.line = item.meta.line
            literal.column = item.meta.column
         elif item.data=='json_object_resource':
            literal.value = item.children[0].value if len(item.children)>0 else ''
            literal.type = LiteralType.JSON_OBJECT
            literal.line = item.meta.line
            literal.column = item.meta.column
         elif item.data=='json_array_resource':
            literal.value = item.children[0].value if len(item.children)>0 else ''
            literal.type = LiteralType.JSON_ARRAY
            literal.line = item.meta.line
            literal.column = item.meta.column
         elif item.data=='parameter_literal':
            literal = None
         elif item.data=='empty_parameter':
            literal = ParameterLiteral(None)
            literal.line = item.meta.line
            literal.column = item.meta.column
         elif item.data=='yaml_parameter':
            literal = ParameterLiteral(item.children[0].value if len(item.children)>0 else '',LiteralType.YAML)
            literal.line = item.meta.line
            literal.column = item.meta.column
         elif item.data=='json_object_parameter':
            literal = ParameterLiteral(item.children[0].value if len(item.children)>0 else '',LiteralType.JSON_OBJECT)
            literal.line = item.meta.line
            literal.column = item.meta.column
         elif item.data=='json_array_parameter':
            literal = ParameterLiteral(item.children[0].value if len(item.children)>0 else '',LiteralType.JSON_ARRAY)
            literal.line = item.meta.line
            literal.column = item.meta.column
         elif item.data=='task_list':
            assert len(item.children)==1, 'meet shorthand not supported'
         elif item.data=='task':
            index += 1
            step = Task(index,item.children[0].value,merge=merge)
            step.line = item.children[0].line
            step.column = item.children[0].column
            if iterate:
               step = Iterate(step)
               step.index = step.step.index
            realize_step(step)
            subject = step
         elif item.data=='destination':
            if isinstance(item.children[-1],Token) and item.children[-1].type=='LABEL':
               statement.destination = item.children[-1].value[1:]
               statement.merge_destination = isinstance(item.children[0],Token) and item.children[0].type=='MERGE'
            elif isinstance(item.children[-1],Tree) and item.children[-1].data.value=='resource':
               index += 1
               subject = ResourceSink(index,merge=isinstance(item.children[0],Token) and item.children[0].type=='MERGE')
               realize_step(subject)
         elif item.data=='doc_comment':
            subject.doc = item.children[0].value
         elif item.data=='declaration':
            kind = item.children[0].value[1:]
            decl = Declaration(kind,item.children[1].value)
            if len(item.children)>2 and isinstance(item.children[2],Token) and item.children[2].type=='EQUAL':
               if kind=='flow':
                  assert False, f'{item.data.line}:{item.data.column} flows can not have base types'
               decl.base = item.children[3].value
            if kind=='flow':
               workflow.name = decl.name
            key = (kind,decl.name)
            if key in workflow.declarations:
               # warning needed, last one wins
               pass
            workflow.declarations[key] = decl
            subject = decl
         else:
            line = item.meta.line
            column = item.meta.column
            logging.error(item.data)
            logging.error(str(item))
            assert False, f'{line}:{column} Cannot handle item: {item.data}'

      index += 1
      end.index = index
      workflow.indexed.append(end)

      for index,item in enumerate(workflow.indexed):
         assert index==item.index, f'Index for item at {index} does not match: {item}'

      return workflow



import sys
import os
import json

import click

from littleflow import __version__, Parser, Compiler, Runner, Context, graph

@click.group()
def cli():
   pass

@cli.command()
def version():
   print('.'.join(map(str,__version__)))

@cli.command()
@click.argument('files',nargs=-1)
def compile(files):
   for file in files:
      p = Parser()
      c = Compiler()
      try:
         with open(file,'r') as input:
            model = p.parse(input)
            print('Model:')
            print(model)
            print()
            flow = c.compile(model)
            if flow.name is not None:
               print(f'Flow: {flow.name}')
            print(flow.F)
            for index in range(len(flow)):
               print(flow[index])

      except FileNotFoundError as ex:
         print(f'Cannot open {file}',file=sys.stderr)

class LogContext(Context):

   def __init__(self,flow,state=None,activation=None,show_cache=False):
      super().__init__(flow,state=state,activation=activation)
      self.show_cache = show_cache

   def input_for(self,index):
      if self.show_cache:
         print(self.cache)
      return super().input_for(index)

   def ended(self,value):
      print(f'output → {value}')

   def output_for(self,index,value):
      print(f'{index} → {value}')
      super().output_for(index,value)

   def input_for(self,index):
      value = super().input_for(index)
      print(f'{value} → {index}')
      return value

   def end(self,tasks):
      print('E',str(self.S.flatten()),str(self.A.flatten()),str((1*tasks).flatten()),str(self.T.flatten()))
      super().end(tasks)

   def start(self,tasks):
      E = 1*tasks
      print('S',str(self.S.flatten()),str(self.A.flatten()),str(E.flatten()),str(self.T.flatten()))
      super().start(tasks)

   def start_task(self,invocation,input):
      print(f'{input} → {invocation.index} ({invocation.name})')
      self.output_for(invocation.index,input)
      super().start_task(invocation,input)

@cli.command()
@click.option('--limit',type=int,default=-1,help='Iteration limit')
@click.option('--show-cache',is_flag=True)
@click.option('--input',help='The workflow input')
@click.argument('workflow')
def run(limit,show_cache,input,workflow):

   if input is not None:
      if input[0]=='@':
         with open(input[1:],'r') as data:
            input = json.load(data)
      else:
         input = json.loads(input)

   p = Parser()
   c = Compiler()
   try:
      with open(workflow,'r') as raw:
         model = p.parse(raw)
         flow = c.compile(model)
   except FileNotFoundError as ex:
      print(f'Cannot open {file}',file=sys.stderr)

   runner = Runner()
   context = LogContext(flow)
   runner.start(context,input)
   count = 0
   while (limit<0 or count<limit) and not context.ending.empty():
      count += 1
      runner.next(context,context.ending.get())

@cli.command()
@click.option('--no-docs',is_flag=True)
@click.option('--vertical',is_flag=True)
@click.argument('workflow')
def doc(no_docs,vertical,workflow):
   p = Parser()
   c = Compiler()
   try:
      with open(workflow,'r') as input:
         model = p.parse(input)
         flow = c.compile(model)
   except FileNotFoundError as ex:
      print(f'Cannot open {workflow}',file=sys.stderr)

   graph(flow,sys.stdout,embed_docs=not no_docs,left_to_right=not vertical)


if __name__=='__main__':
   cli()



import json

import yaml

from .model import Workflow, SubFlow, Task, LiteralSource, Start, End, LiteralType
from .flow import Flow, Source, Sink, InvokeTask, InvokeFlow, StartFlow

def compile_literal(text,type=LiteralType.EMPTY):
   if type==LiteralType.EMPTY:
      return {}

   text = text.strip()

   if type==LiteralType.YAML:
      if len(text)==0:
         return {}
      try:
         value = yaml.load(text,Loader=yaml.Loader)
         return value
      except yaml.YAMLError as ex:
         raise ValueError(f'Cannot parse YAML literal: {ex}')
   elif type==LiteralType.JSON_ARRAY:
      if len(text)==0:
         return []
      try:
         value = json.loads('[' + text + ']')
         return value
      except json.JSONDecodeError as ex:
         raise ValueError(f'Cannot parse JSON array literal: {ex}')

   elif type==LiteralType.JSON_OBJECT:
      if len(text)==0:
         return {}
      try:
         value = json.loads('{' + text + '}')
         return value
      except json.JSONDecodeError as ex:
         raise ValueError(f'Cannot parse JSON object literal: {ex}')


class Compiler:

   def __init__(self):
      pass

   def compile(self,model):

      size = len(model.indexed)
      assert size>2, 'The workflow contains no flows.'
      flow = Flow(size,name=model.name)

      for index, step in enumerate(model.indexed):
         if isinstance(step,Task):
            decl = model.declarations.get(('task',step.name))
            value = {}
            if decl is not None:
               if decl.parameters is not None:
                  try:
                     value = compile_literal(decl.parameters.value,decl.parameters.type)
                  except ValueError as ex:
                     raise ValueError(f'{decl.parameters.line}:{decl.parameters.column} {ex}')
            if step.parameters is not None:
               try:
                  invocation_value = compile_literal(step.parameters.value,step.parameters.type)
                  if len(value)==0:
                     value = invocation_value
                  elif type(value)!=type(invocation_value):
                     raise ValueError(f'{step.parameters.line}:{step.parameters.column} the type of the declaration parameters ({type(value)}) does not match the type of the invocation parameters ({type(invocation_value)})')
                  elif type(value)==list:
                     value = invocation_value
                  elif type(value)==dict:
                     for key, value in invocation_value.items():
                        value[key] = value
                  else:
                     value = invocation_value

               except ValueError as ex:
                  raise ValueError(f'{step.parameters.line}:{step.parameters.column} {ex}')
            flow[index] = InvokeTask(index,step.name,value,merge=step.merge)
            if decl is not None:
               flow[index].doc = decl.doc
               flow[index].base = decl.base
         elif isinstance(step,LiteralSource):
            try:
               value = compile_literal(step.value,step.type)
               flow[index] = Source(index,value)
            except ValueError as ex:
               raise ValueError(f'{step.line}:{step.column} {ex}')
         elif isinstance(step,SubFlow):
            flow[index] = InvokeFlow(index,merge=step.merge)
         elif isinstance(step,Start):
            flow[index] = StartFlow(index)
         elif isinstance(step,End):
            flow[index] = Sink(index)
         else:
            raise NotImplementedError(f'Support for {step.__class__.__name__} not implemented')

      flows = [(0,model.flows[0])]
      while len(flows)>0:
         prior, subflow = flows.pop()
         for statement in subflow.statements:
            current = prior
            if statement.source is not None:
               named = subflow.named_outputs.get(statement.source)
               if named is None:
                  raise ValueError(f'Unknown output label {statement.source} at {statement.line}:{statement.column}: {statement}')
               current = named.index

            for step in statement.steps:
               assert flow.F[current,step.index]==0, f'Transition from {current}->{step.index} already exists.'
               flow.F[current,step.index] = 1
               if isinstance(step,SubFlow):
                  flows.append((step.index,step))
                  current = step.named_inputs.get('end').index
               else:
                  current = step.index

            if statement.destination is not None:
               named = subflow.named_inputs.get(statement.destination)
               if named is None:
                  raise ValueError(f'Unknown input label {statement.destination} at {statement.line}:{statement.column}: {statement}')
               flow[named.index].merge = flow[named.index].merge or statement.merge_destination
               flow.F[current,named.index] = 1
            else:
               named = subflow.named_inputs.get('end')
               assert named is not None, 'The end destination is missing for the subflow'
               flow.F[current,named.index] = 1
         # fixup for inputs and outputs within the flow
         for name, source in subflow.named_outputs.items():
            if name=='start':
               continue
            target = subflow.named_inputs.get(name)
            if target is not None:
               flow.F[source.index,target.index] = 1
      return flow



from queue import SimpleQueue
import types

import numpy as np
from .flow import Flow, Source, Sink, InvokeTask, InvokeFlow, StartFlow

class TaskContext:

   def invoke(self,context,invocation,input):
      immediate = np.zeros(context.S.shape,dtype=int)
      immediate[invocation.index] = 1
      context.ending.put(immediate)

def pass_parameters(func):
   if hasattr(func,'__invocation__'):
      getattr(func,'__invocation__').append('raw_parameters')
   else:
      setattr(func,'__invocation__',['raw_parameters'])
   return func

def pass_input(func):
   if hasattr(func,'__invocation__'):
      getattr(func,'__invocation__').append('input')
   else:
      setattr(func,'__invocation__',['input'])
   return func

def task(registry=None,apply=True):
   def register(func):
      if hasattr(func,'__invocation__'):
         getattr(func,'__invocation__').append('apply')
      else:
         setattr(func,'__invocation__',['apply'])
      if registry is not None:
         registry[func.__name__] = func
      return func
   return register

class FunctionTaskContext(TaskContext):

   def __init__(self,lookup):
      self._lookup = lookup

   def invoke(self,context,invocation,input):
      f = self._lookup.get(invocation.name)
      if f is None:
         raise ValueError(f'Cannot find task {invocation.name} in function lookup.')
      if not isinstance(f,types.FunctionType):
         raise ValueError(f'Task {invocation.name} resolves to non-function')
      options = getattr(f,'__invocation__') if hasattr(f,'__invocation__') else []
      if 'apply' in options:
         match input:
            case list():
               args = input
            case tuple():
               args = list(input)
            case _:
               args = [input]
      else:
         args = [input] if 'input' in options else [{}]
      keywords = {}
      if invocation.parameters is not None:
         if 'raw_parameters' not in options:
            if type(invocation.parameters)==dict:
               for key,value in invocation.parameters.items():
                  keywords[key] = value
            elif type(invocation.parameters)==list:
               args += invocation.parameters
         else:
            args.append(invocation.parameters)

      output = f(*args,**keywords)

      if output is None:
         output = {}
      elif type(output)==tuple:
         output = list(output)
      elif type(output)!=dict and type(output)!=list:
         output = [output]

      context.output_for(invocation.index,output)
      immediate = np.zeros(context.S.shape,dtype=int)
      immediate[invocation.index] = 1
      context.ending.put(immediate)

def merge(input):
   if type(input)!=list:
      return input
   result = {}
   for item in input:
      if type(item)!=dict:
         raise ValueError(f'Item in merge is not of type dict: {type(item)}')
      for key, value in item.items():
         result[key] = value
   return result

class Context:
   def __init__(self,flow,state=None,activation=None,cache={},task_context=TaskContext()):
      self._flow = flow
      self._A = activation if activation is not None else np.zeros((self.F.shape[0],1),dtype=int)
      self._T = flow.F.sum(axis=0)
      self._T[0] = 1
      self._T = self._T.reshape((self.F.shape[0],1))
      self._S = state if state is not None else np.zeros((self.F.shape[0],1),dtype=int)
      self._ends = SimpleQueue()
      self._cache = cache
      self._task_context = task_context


   @property
   def flow(self):
      """
      The workflow for the context
      """
      return self._flow

   @property
   def F(self):
      """
      The step transition matrix.
      """
      return self._flow.F

   @property
   def S(self):
      """
      Indicates which steps are currently active.
      """
      return self._S

   @property
   def A(self):
      """
      The currently accumulated activations
      """
      return self._A

   @property
   def T(self):
      """
      The activation threshold
      """
      return self._T

   @property
   def ending(self):
      return self._ends

   @property
   def cache(self):
      return self._cache

   @property
   def task_context(self):
      return self._task_context

   @task_context.setter
   def task_context(self,value):
      self._task_context = value

   def new_transition(self):
      return np.zeros((self.F.shape[0],1),dtype=int)

   def start(self,tasks):
      """
      Called when steps are started. The tasks argument is a boolean vector
      whose position correspond to the indexed steps that should be started.
      The default implementation immediately ends tasks.
      """
      immediate = np.zeros(tasks.shape,dtype=int)
      for index,task in enumerate(tasks.flatten()):
         if task:
            invocation = self.flow[index]
            input = self.input_for(index)
            assert input is not None, f'None value return for {index}'
            if isinstance(invocation,InvokeTask):
               if invocation.merge:
                  input = merge(input)
               self.start_task(invocation,input)
            elif isinstance(invocation,Source):
               self.output_for(index,invocation.value)
               immediate[index] = 1
            elif isinstance(invocation,Sink):
               if invocation.merge:
                  input = merge(input)
               self.output_for(index,input)
               immediate[index] = 1
            elif isinstance(invocation,InvokeFlow):
               if invocation.merge:
                  input = merge(input)
               self.output_for(index,input)
               immediate[index] = 1
            elif isinstance(invocation,StartFlow):
               self.output_for(index,input)
               immediate[index] = 1
            else:
               immediate[index] = 1

      if immediate.sum()>0:
         self.ending.put(immediate)

   def end(self,tasks):
      """
      Called when steps end. The tasks argument is a boolean vector
      whose position correspond to the indexed steps that have finished.
      """
      pass

   def ended(self,value):
      pass

   def output_for(self,index,value):
      if value is None:
         return
      if type(value)!=list and type(value)!=dict:
         raise ValueError(f'The value of the input must be a list or dict: {type(value)}')
      self._cache[index] = value
      if index==(self.F.shape[0]-1):
         self.ended(value)

   def input_for(self,index):
      if index == 0:
         item = self._cache.get(-1)
         return {} if item is None else item
      value = []
      for target, transition in enumerate(self.F.T[index].flatten()):
         if transition > 0:
            item = self._cache.get(target)
            if item is not None:
               value.append(item)
      if len(value)==0:
         return {}
      elif len(value)==1:
         return value[0]
      else:
         return value

   def start_task(self,invocation,input):
      self._task_context.invoke(self,invocation,input)

   def accumulate(self,value):
      pass


class Runner:

   def __init__(self):
      pass

   def start(self,context,input=None):
      if input is not None:
         context.output_for(-1,input)
      context.A[0] = 1
      context.ending.put(context.new_transition())
      context.accumulate(context.A)


   def next(self,context,E):
      """
      Runs the algorithm forward from a vector representing steps that have ended.

      context - The flow context
      E - a zero/one or boolean vector indicating which steps have ended
      """
      # allow array of booleans
      E = 1*E
      activations = context.F.T.dot(E)
      context.accumulate(activations)
      context._A = context._A + activations
      if E.sum()>0:
         context.end(E>0)
      # TODO: should we allow more than zero/one vectorss
      N = context.A >= context.T
      context._A = context.A - 1*N * context.T
      context._S = context.S - E + 1*N
      # Guarantee positive semi-definite
      context._S = context._S + np.multiply(context._S,-1*(context._S<0))
      if (1*N).sum()>0:
         context.start(N)
      return context.S.sum()>0



from dataclasses import dataclass, field
from enum import Enum

class LiteralType(Enum):
   EMPTY = 0
   YAML = 1
   JSON_OBJECT = 2
   JSON_ARRAY = 3

@dataclass
class ParameterLiteral:
   value : str
   type : LiteralType = LiteralType.EMPTY
   line : int = 0
   column : int = 0

@dataclass
class Destination:
   pass

@dataclass
class LabelDestination(Destination):
   name : str

@dataclass
class Step:
   pass

@dataclass
class Start(Step):
   index : int = 0

@dataclass
class End(Step):
   index : int

@dataclass
class Iterate(Step):
   step : Step = None
   index : int = -1

@dataclass
class LiteralSource(Step):
   index : int
   value : str
   type : LiteralType = LiteralType.EMPTY
   line : int = 0
   column : int = 0

@dataclass
class ResourceSource(Step):
   index : int
   uri : str = ''
   parameters: ParameterLiteral = None

@dataclass
class ResourceSink(Step):
   index : int
   uri : str = ''
   merge : bool = False
   parameters: ParameterLiteral = None

@dataclass
class Task(Step):
   index : int
   name : str
   merge : bool = False
   line : int = 0
   column : int = 0
   parameters : ParameterLiteral = None

@dataclass
class Statement:
   source : str = None
   destination : str = None
   merge_destination : bool = False
   steps : list[Step] = field(default_factory=list)
   line : int = 0
   column : int = 0

@dataclass
class SubFlow(Step):
   index : int
   merge : bool = False
   named_inputs : dict = field(default_factory=dict)
   named_outputs : dict = field(default_factory=dict)
   statements : list[Statement] = field(default_factory=list)

@dataclass
class Declaration:
   kind : str
   name : str
   base : str = None
   doc : str = None
   parameters : ParameterLiteral = None

@dataclass
class Workflow:
   name : str = None
   indexed : list[Step] = field(default_factory=list)
   flows : list[SubFlow] = field(default_factory=list)
   declarations : dict = field(default_factory=dict)
   doc : str = None



from .flow import StartFlow, Source, Sink, InvokeTask, InvokeFlow

def mangle(name):
   return name.replace(':','_COLON_').replace('-','_HYPHEN_')
def graph_name(obj,end=-1):
   if isinstance(obj,InvokeTask):
      return f'{mangle(obj.name)}.{obj.index}'
   elif isinstance(obj,StartFlow):
      return '[*]'
   elif isinstance(obj,Sink):
      return '[*]' if obj.index==end else f'_end_{obj.index}_'
   elif isinstance(obj,InvokeFlow):
      return f'_start_{obj.index}_'
   else:
      return None

def shortdesc(doc):
   oneliner, _, rest = doc.strip().partition('\n')
   return oneliner

def graph(flow,output,format='mermaid',name='workflow',embed_docs=True,left_to_right=True):
   assert format=='mermaid', 'Only mermaid is currently supported'
   print(f'stateDiagram-v2',file=output)
   if left_to_right:
      print('  direction LR',file=output)
   size = len(flow)
   for index in range(size):
      invocation = flow[index]
      name = graph_name(invocation,end=size-1)
      if isinstance(invocation,InvokeFlow):
         print(f'  state {name} <<fork>>',file=output)
      elif isinstance(invocation,Sink) and index!=(size-1):
         print(f'  state {name} <<join>>',file=output)
      elif isinstance(invocation,InvokeTask):
         print(f'  state "{invocation.name}" as {name}',file=output)

   for index,row in enumerate(flow.F):
      source = flow[index]
      source_name = graph_name(source,end=size-1)
      for target_index, value in enumerate(row.flatten()):
         if value==0:
            continue
         target = flow[target_index]
         #print(source,target)
         target_name = graph_name(target,end=size-1)
         print(f'  {source_name}-->{target_name}',file=output)
         if  embed_docs and isinstance(target,InvokeTask) and target.doc is not None:
            print(f'  note left of {target_name}',file=output)
            print(f'    {shortdesc(target.doc)}',file=output)
            print(f'  end note',file=output)



__version__=(0,12,0)
__author__='Alex Miłowski'
__author_email__='alex@milowski.com'
from .parser import Parser
from .model import Workflow, Declaration, SubFlow, Statement, Start, End, Iterate, Task, LiteralSource, ResourceSource, ResourceSink, ParameterLiteral
from .compiler import Compiler
from .runner import Context, Runner, TaskContext, FunctionTaskContext, pass_input, pass_parameters, merge, task
from .flow import Invocation, Source, Sink, InvokeTask, InvokeFlow, Flow
from .doc import graph_name, graph
from .utils import run_workflow, TaskFailure, WorkflowFailure
